/*
  unsorttable: shuffle rows of data in a NoSQL table.

  Original code from 'unsort' v1.0, by Agustn Ferrin .
  Source: http://gatopeichs.pbwiki.com/#unsortthemissingUNIXfilter

  Included in NoSQL under the GPL with permission of the author.

  Copyright (c) 2008 Carlo Strozzi

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 dated June, 1991.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef HELPDIR
#define HELPDIR "/usr/local/nosql/help"
#endif
#ifndef HELPFILE
#define HELPFILE (HELPDIR "/unsorttable.txt")
#endif

#ifndef DOCDIR
#define DOCDIR "/usr/local/nosql/doc"
#endif
#ifndef COPYING
#define COPYING (DOCDIR "/COPYING")
#endif
#ifndef WARRANTY
#define WARRANTY (DOCDIR "/WARRANTY")
#endif

#define HEADER \
  "'unsort', the missing unix filter.\n" \
  "V1.0 written by gatopeich in 2006 (see http://gatopeichs.pbwiki.com/).\n" \
  "Built on " __DATE__ " " __TIME__ ".\n"

/* unsort unix filter */
/* Each input line is assigned a random integer, and sorted based on it */
/* A binary tree is used, taking advantage of the randomness of indexes, */
/* which should keep the tree balanced at no cost. */

/* My sincere thanks to folkert@vanheusden.com for his original 'unsort'
   that came so handy until I found it buggy :-) */

char usage[] = "usage: %s [OPTION]..." "\n"
"Randomly shuffle rows from standard input and write to standard output." "\n"
"    -s SEED        use SEED for random number generation, otherwise use time." "\n"
"    -v             verbose run (print seed & stats on stderr)" "\n"
"    -h             display this help and exit" "\n"
"    -V             output version information and exit" "\n"
"    -N             do not print the output table header" "\n";
char version[] = HEADER ;

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

typedef struct node_st
{
    long int index;
    char *line;
    size_t line_length;

    struct node_st *left;
    struct node_st *right;
} node;

void write_node(node *n)
{
    if( 1 != fwrite( n->line, n->line_length, 1, stdout) )
    {
        fprintf( stderr, "fwrite error.\n");
        exit(1);
    }
}

/* Some stats to check how balanced is our random tree */
static int jumps_l=0, jumps_r=0;

int insert_node ( node **root, node *n)
{
    while ( NULL != *root )
    {
        if ( (*root)->index < n->index )
        {
            root = &(*root)->right;
            jumps_r++;
        }
        else if ( (*root)->index > n->index )
        {
            root = &(*root)->left;
            jumps_l++;
        }
        else
        {
            fprintf( stderr, "Ops!, index %ld was repeated!!!\n", n->index );
            return 0;
        }
    }

    //~ fprintf( stderr, "Adding index %ld...\n", n->index );
    *root = n;

    return 1;
}

void write_and_free_nodes (node *n)
{
    if ( NULL == n )
    {
        return;
    }
    else
    {
        write_and_free_nodes ( n->left );
        write_node(n);
        node *r = n->right;
        free(n);
        write_and_free_nodes ( r );
        return;
    }
}

int main(int argc, char **argv)
{
    int quiet = 1, no_hdr = 0;
    node *list_start=NULL;

    /* Prepare a time-based seed */
    struct timeval tv;
    gettimeofday( &tv, NULL);
    long int r = tv.tv_sec * 1000000 + tv.tv_usec;

    int i=1;
    while( i < argc )
    {
        if( argv[i][0] == '-' )
        {
            int j,o; j=1;
            while( (o=argv[i][j++]) )
            {
                switch(o)
                {
                case 's':
                    if( argv[i][j] || argc <= ++i) goto badargs;
                    r = atol(argv[i]);
                    break;
                case 'v':
                    quiet = 0;
                    break;
                case 'N':
                    no_hdr = 1;
                    break;
                case 'h':
                    printf( usage, argv[0] );
                    exit(0);
                case 'V':
                    printf( version );
                    exit(0);
                default:
                badargs:
                    printf("%s: bad arguments.\n"
                        "Try '%s -h' for more information.\n", argv[0], argv[0]);
                    exit(1);
                }
            }
        }
        i++;
    }

    if(!quiet) fprintf(stderr, "Using seed: %ld\n", r);
    srandom(r);

    node *n = (node *) calloc(1, sizeof(node));
    ssize_t l;
    
    /* Read table headers line, and output it if needed */
    if( -1 != (l=getline( &n->line, &n->line_length, stdin)) )
    {
    	if(!no_hdr) {
            n->line_length = l;
    		write_node(n);
		}
	}
    
    while( -1 != (l=getline( &n->line, &n->line_length, stdin)) )
    {
        n->line_length = l;
        do
        {
            n->index = random();
        } while( !insert_node( &list_start, n) );
        n = (node *) calloc(1, sizeof(node));
    }
    free(n); /* last is unused! */

    write_and_free_nodes(list_start);

    if(!quiet)
        fprintf( stderr, "Jumps: %d left, %d right, balance: %f\n",
            jumps_l, jumps_r,
            ((jumps_l > jumps_r) ? ((double)jumps_r/jumps_l) : ((double)jumps_l/jumps_r)) );

    return 0;
}

/* EOF */
